// gluggMoreShapes
// Creates a few standard shapes not covered in gluggShapesDemo:
// Sphere, Cone, Cylinder and Ellipsoid
// By Ingemar 2021.

// 210603: Cleaned up some really unnecessary warnings.
// Corrected the ellipsoid which only worked with height 1!
// 211224: New version using gluggShapes.
// 2023: C++ version

// uses framework Cocoa
// uses framework OpenGL
#define MAIN
#include <stdlib.h>
#include "MicroGlut.h"
#include "GL_utilities.h"
#include "VectorUtils4.h"
#include "glugg.h"
#include "gluggShapes.h"

#ifndef min
#define min(a,b) (((a)<(b))?(a):(b))
#define max(a,b) (((a)>(b))?(a):(b))
#endif

GLuint program; // Shader
mat4 projectionMatrix;

static GLfloat red[4] = { 0.8, 0.1, 0.0, 1.0 };
static GLfloat green[4] = { 0.0, 0.8, 0.2, 1.0 };
static GLfloat blue[4] = { 0.2, 0.2, 1.0, 1.0 };
static GLfloat yellow[4] = { 0.8, 0.8, 0.0, 1.0 };
//static GLfloat magenta[4] = { 1.0, 0.2, 1.0, 1.0 };
//static GLfloat cyan[4] = { 0, 1, 1, 1.0 };

gluggModel s1, s2, s3, s4;
GLuint s0;
int count0;

void display(void)
{
	float time = glutGet(GLUT_ELAPSED_TIME) / 1000.0;

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    mat4 cam = lookAt(0, 2, 8, 0, 0, 0, 0,1,0);
	
	mat4 m;
    m = cam * T(2.0, 2.0,0.0) * Ry(time*2) * Rx(3*sin(time/5));
    glUniformMatrix4fv(glGetUniformLocation(program, "modelviewMatrix"), 1, GL_TRUE, m.m);
	glUniform4fv(glGetUniformLocation(program, "color"), 1, red);
	gluggDrawModel(s1, program);

    m = cam * T(-2.0, -2.0,0.0) * Ry(time*2) * Rx(5*sin(time/2));
    glUniformMatrix4fv(glGetUniformLocation(program, "modelviewMatrix"), 1, GL_TRUE, m.m);
	glUniform4fv(glGetUniformLocation(program, "color"), 1, blue);
	gluggDrawModel(s2, program);

    m = cam * T(-2.0, 1.4,0.0) * Ry(time*2) * Rx(2*sin(time/5));
    glUniformMatrix4fv(glGetUniformLocation(program, "modelviewMatrix"), 1, GL_TRUE, m.m);
	glUniform4fv(glGetUniformLocation(program, "color"), 1, green);
	gluggDrawModel(s3, program);

    m = cam * T(2.0, -2.0,0.0) * Ry(time*2) * Rx(3*sin(time/5));
    glUniformMatrix4fv(glGetUniformLocation(program, "modelviewMatrix"), 1, GL_TRUE, m.m);
	glUniform4fv(glGetUniformLocation(program, "color"), 1, yellow);
	gluggDrawModel(s4, program);

    glutSwapBuffers();
}

void myinit()
{
    glEnable(GL_DEPTH_TEST); /* enable z buffer */
    glDisable(GL_CULL_FACE);

	program = loadShaders("sphere.vert", "sphere.frag");
	
	gluggBegin(0);
	gluggSphere(15, 10, 0.75);
	s1 = gluggBuildModel(0);
	gluggBegin(0);
	gluggCone(20, 2.0, 1.0);
	s2 = gluggBuildModel(0);
	gluggBegin(0);
	gluggCylinder(20, 1.5, 1.0);
	s3 = gluggBuildModel(0);
	gluggBegin(0);
	gluggEllipsoid(15, 14, SetVector(0.75, 2, 1.5));
	s4 = gluggBuildModel(0);

	projectionMatrix = frustum(-0.5, 0.5, -0.5, 0.5, 1.0, 60.0);

	glUniformMatrix4fv(glGetUniformLocation(program, "projectionMatrix"), 1, GL_TRUE, projectionMatrix.m);
	
    glClearColor (1.0, 1.0, 1.0, 1.0);
}


int main(int argc, char **argv)
{
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_SINGLE | GLUT_RGB | GLUT_DEPTH);
    glutInitContextVersion(3, 2);
    glutInitWindowSize(500, 500);
    glutCreateWindow("GLUGG more shapes");
    glutRepeatingTimer(15);
    myinit();
    glutDisplayFunc(display);
    glutMainLoop();
}
