// Sweep a 2D curve around an axis.
// Either from discrete points or by using a Catmull-Rom spline.
// Possible variant: Sweep around a curve, e.g. two 2D curves.
// Another variant: Take a backwards turn a step in to produce an inside.

// First version 2021. I am not sure all normals are correct. Non-smooth are not closed here.
// But I can leave something for you to figure out, can't I?
// Modified for GLUGG 1.3 2022.

// uses framework Cocoa
// uses framework OpenGL
#define MAIN
#include <stdlib.h>
#include "MicroGlut.h"
#include "GL_utilities.h"
#include "VectorUtils4.h"
#include "glugg.h"

#ifndef min
#define min(a,b) (((a)<(b))?(a):(b))
#define max(a,b) (((a)>(b))?(a):(b))
#endif

GLuint program; // Shader

vec3 CatmullRom(vec3 p1, vec3 p2, vec3 p3, vec3 p4, float u)
{
	vec3 p = 
				p1 * (-u*u*u/2 + u*u -u/2) +
				p2 * (3*u*u*u/2 - 5*u*u/2 + 1) +
				p3 * (-3*u*u*u/2 + 2*u*u + u/2) +
				p4 * (u*u*u/2 - u*u/2);
	return p;
};

vec3 CatmullRomDerivative(vec3 p1, vec3 p2, vec3 p3, vec3 p4, float u)
{
	vec3 p =
				p1 * (-3*u*u/2 + 2*u -0.5) + // Note! Do NOT write 1/2!
				p2 * (9*u*u/2 - 5*u) +
				p3 * (-9*u*u/2 + 4.*u + 0.5) +
				p4 * (3*u*u/2 - u);
	return p;
};

// Look up interpolation with Catmull-Rom
gluggModel smoothsweep(vec3 *points, int pointcount, int xsteps, int ysteps)
{
	gluggBegin(GLUGG_TRIANGLE_STRIP);

	for (int j = 0; j < ysteps; j++)
	{
		float b = (float)j*pointcount / ysteps;
//		float c = (float)((j-1+pointcount) % pointcount)*pointcount / ysteps;
		float c = (j-1.0)*pointcount / ysteps;
		if (c < 0) c = c + pointcount; // This looks wrong so why does it work???
	
		gluggMode(GLUGG_TRIANGLE_STRIP);
		for (int i = 0; i <= xsteps; i++)
		{
			float a1 = i * 2.0 * M_PI / xsteps; // xsteps = rotational steps
//			float a0 = (i-1) * 2.0 * M_PI / xsteps;
						
	// Tv punkter lngs kanten. Jag mste hitta mina 4 kontrollpunkter tv gnger eftersom de kan g "ver kanten"
			vec3 p1 = points[((int)b) % pointcount];
			vec3 p2 = points[((int)b+1) % pointcount];
			vec3 p3 = points[((int)b+2) % pointcount];
			vec3 p4 = points[((int)b+3) % pointcount];

			vec3 e2 = CatmullRom(p1, p2, p3, p4, b - (int)b);
			vec3 d1 = CatmullRomDerivative(p1, p2, p3, p4, b - (int)b);
			d1 = SetVector(d1.y, -d1.x, 0);

			p1 = points[((int)c) % pointcount];
			p2 = points[((int)c+1) % pointcount];
			p3 = points[((int)c+2) % pointcount];
			p4 = points[((int)c+3) % pointcount];

			vec3 e3 = CatmullRom(p1, p2, p3, p4, c - (int)c);
			vec3 d2 = CatmullRomDerivative(p1, p2, p3, p4, c - (int)c);
			d2 = SetVector(d2.y, -d2.x, 0);


			// Calculate 2 points along same slice
			
			p2 = SetVector(e2.x * cos(a1), e2.y, e2.x * sin(a1));
			p3 = SetVector(e3.x * cos(a1), e3.y, e3.x * sin(a1));

			vec3 pn2 = SetVector(d1.x * cos(a1), d1.y, d1.x * sin(a1));
			vec3 pn3 = SetVector(d2.x * cos(a1), d2.y, d2.x * sin(a1));

	// Done making points and normals. Now create polygons!
			gluggNormalv(pn3);
		    gluggVertexv(p3);
			gluggNormalv(pn2);
		    gluggVertexv(p2);
		}
	}

	return gluggBuildModel(0);
}

gluggModel sweep(vec3 *points, int pointcount, int steps)
{
	gluggBegin(GLUGG_TRIANGLE_STRIP);

	for (int j = 0; j < pointcount-1; j++)
	{
		gluggMode(GLUGG_TRIANGLE_STRIP);
		vec3 e1 = points[max(0, j-1)];
		vec3 e2 = points[j];
		vec3 e3 = points[min(j+1, pointcount-1)];
		vec3 e4 = points[min(j+2, pointcount-1)];
		
		for (int i = 0; i <= steps; i++)
		{
			float a1 = i * 2.0 * M_PI / steps;
			
			// Calculate 4 points along same slice
			
			vec3 p1 = SetVector(e1.x * cos(a1), e1.y, e1.x * sin(a1));
			vec3 p2 = SetVector(e2.x * cos(a1), e2.y, e2.x * sin(a1));
			vec3 p3 = SetVector(e3.x * cos(a1), e3.y, e3.x * sin(a1));
			vec3 p4 = SetVector(e4.x * cos(a1), e4.y, e4.x * sin(a1));

			vec3 pn2 = normalize(VectorAdd(p1, p3));
			vec3 pn3 = normalize(VectorAdd(p2, p4));

	// Done making points and normals. Now create polygons!
			gluggNormalv(pn2);
		    gluggVertexv(p2);
			gluggNormalv(pn3);
		    gluggVertexv(p3);
		}
	}

	return gluggBuildModel(0);
}

// X is the shape, Y should vary monotonously, Z is ignored
vec3 testdata1[] =
{
		vec3(0,0,0),
		vec3(0.4,0.4,0),
		vec3(0.2,0.8,0),
		vec3(1.0,1.2,0),
		vec3(0.6,1.6,0),
		vec3(1.4,2.0,0),
		vec3(0.0,2.0,0),
};

vec3 testdata2[] =
{
		vec3(0,0,0),
		vec3(0.2,0.4,0),
		vec3(0.3,0.8,0),
		vec3(0.2,1.2,0),
		vec3(0.3,1.6,0),
		vec3(0.5,2.0,0),
		vec3(0.0,2.0,0)
};

vec3 testdata3[] =
{
		vec3(0,0,0),
		vec3(0.8,0.2,0),
		vec3(0.5,0.4,0),
		vec3(0.4,0.6,0),
		vec3(0.3,0.8,0),
		vec3(0.7,1.0,0),
		vec3(0.0,1.0,0),
};

static GLfloat red[4] = { 0.8, 0.1, 0.0, 1.0 };
static GLfloat green[4] = { 0.0, 0.8, 0.2, 1.0 };
static GLfloat blue[4] = { 0.2, 0.2, 1.0, 1.0 };
//static GLfloat yellow[4] = { 1, 1, 0.2, 1.0 };
//static GLfloat magenta[4] = { 1.0, 0.2, 1.0, 1.0 };
//static GLfloat cyan[4] = { 0, 1, 1, 1.0 };

gluggModel s1, s2, s3;
gluggModel ss1, ss2, ss3;

mat4 projectionMatrix;

int smooth = 1;

void display(void)
{
	float time = glutGet(GLUT_ELAPSED_TIME) / 1000.0;

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    mat4 cam = lookAt(0, 1, 4, 0, 0, 0, 0,1,0);

    mat4 m = cam * T(2.0, -2.5,0.0) * Ry(time*2) * Rx(sin(time/5));
    glUniformMatrix4fv(glGetUniformLocation(program, "modelviewMatrix"), 1, GL_TRUE, m.m);
	glUniform4fv(glGetUniformLocation(program, "color"), 1, red);
	if (smooth)
		gluggDrawModel(ss1, program);
	else
		gluggDrawModel(s1, program);

    m = cam * T(-2.0, -2.5,0.0) * Ry(time*2) * Rx(sin(time/5));
    glUniformMatrix4fv(glGetUniformLocation(program, "modelviewMatrix"), 1, GL_TRUE, m.m);
	glUniform4fv(glGetUniformLocation(program, "color"), 1, blue);
	if (smooth)
		gluggDrawModel(ss2, program);
	else
		gluggDrawModel(s2, program);

    m = cam * T(0.0, 1.5,0.0) * Ry(time*2) * Rx(sin(time/5));
    glUniformMatrix4fv(glGetUniformLocation(program, "modelviewMatrix"), 1, GL_TRUE, m.m);
	glUniform4fv(glGetUniformLocation(program, "color"), 1, green);
	if (smooth)
		gluggDrawModel(ss3, program);
	else
		gluggDrawModel(s3, program);

    glutSwapBuffers();
}

void myinit()
{
    glEnable(GL_DEPTH_TEST); /* enable z buffer */
    glEnable(GL_CULL_FACE);

	program = loadShaders("sweeper.vert", "sweeper.frag");
	
	s1 = sweep(testdata1, 6, 32);
	s2 = sweep(testdata2, 6, 32);
	s3 = sweep(testdata3, 6, 32);

	ss1 = smoothsweep(testdata1, 7, 50, 23);
	ss2 = smoothsweep(testdata2, 7, 50, 23);
	ss3 = smoothsweep(testdata3, 7, 50, 23);
	
	projectionMatrix = frustum(-1.0, 1.0, -1.0, 1.0, 1.0, 60.0);

	glUniformMatrix4fv(glGetUniformLocation(program, "projectionMatrix"), 1, GL_TRUE, projectionMatrix.m);
	
    glClearColor (1.0, 1.0, 1.0, 1.0);
}

void key(unsigned char key, int x, int y)
{
	smooth = !smooth;
}

int main(int argc, char **argv)
{
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_SINGLE | GLUT_RGB | GLUT_DEPTH);
    glutInitContextVersion(3, 2);
    glutInitWindowSize(500, 500);
    glutCreateWindow("GLUGG sweeper");
    glutRepeatingTimer(15);
    glutKeyboardFunc(key);
    myinit();
    glutDisplayFunc(display);
    glutMainLoop();
}
